implementation module Internet.HTTP.TCPIP

import Control.Applicative
import Control.Monad
import Data.Error
import Data.Func
import Data.Functor
import Data.List
import Data.Maybe
import StdEnv
import TCPIP
import Text.URI

import Internet.HTTP

doHTTPRequest :: !HTTPRequest Int !*World -> *(!MaybeErrorString HTTPResponse, !*World)
doHTTPRequest req timeout w
# (ip,w) = lookupIPAddress req.server_name w
| ip =: ?None
	= (Error $ "DNS lookup for " +++ req.server_name +++ " failed.", w)
# (?Just ip) = ip
# (rpt,chan,w) = connectTCP_MT (?Just timeout) (ip, req.server_port) w
| rpt == TR_Expired
	= (Error $ "Connection to " +++ toString ip +++ " timed out.", w)
| rpt == TR_NoSuccess
	= (Error $ "Could not connect to " +++ req.server_name +++ ".", w)
# (?Just {sChannel,rChannel}) = chan
# (rpt,i,sChannel,w) = send_MT (?Just timeout) (toByteSeq req) sChannel w
| rpt <> TR_Success
	= (Error $ "Could not send request to " +++ req.server_name +++ ".", w)
# (rpt,resp,rChannel,w) = receive_MT (?Just timeout) rChannel w
| rpt <> TR_Success
	= (Error $ "Did not receive a reply from " +++ req.server_name +++ ".", w)
# resp = join $ parseResponse <$> toString <$> resp
| isNone resp
	# w = closeChannel sChannel (closeRChannel rChannel w)
	= (Error $ "Server did not respond with HTTP.", w)
# (resp,rChannel,w) = receiveRest (fromJust resp) rChannel w
# w = closeChannel sChannel (closeRChannel rChannel w)
= (resp,w)
where
	receiveRest :: !HTTPResponse !*(TCP_RChannel_ ByteSeq) !*World -> *(!MaybeErrorString HTTPResponse, !*TCP_RChannel_ ByteSeq, !*World)
	receiveRest resp chan w
	# (end,chan,w) = eom chan w
	| end = (Ok resp, chan, w)
	# (rpt,newresp,chan,w) = receive_MT (?Just timeout) chan w
	| rpt <> TR_Success
		# (end,chan,w) = eom chan w
		| end = (Ok resp, chan, w)
		| otherwise = (Error $ req.server_name +++ " hung up during transmission", chan, w)
	| otherwise = case newresp of
		?None
			-> (Error "Internal error in Internet.HTTP.", chan, w)
		?Just newresp
			-> receiveRest {resp & rsp_data=resp.rsp_data +++ toString newresp} chan w

doHTTPRequestFollowRedirects :: !HTTPRequest Int !Int !*World -> *(!MaybeErrorString HTTPResponse, !*World)
doHTTPRequestFollowRedirects req timeout 0 w = (Error "Maximal redirect number exceeded", w)
doHTTPRequestFollowRedirects req timeout maxRedirects w
# (er, w) = doHTTPRequest req timeout w
| isError er = (er, w)
# resp = fromOk er
| isMember resp.HTTPResponse.rsp_code [301, 302, 303, 307, 308]
	= case lookup "location" resp.HTTPResponse.rsp_headers <|> lookup "Location" resp.HTTPResponse.rsp_headers of
		?None = (Error $ "Redirect given but no Location header", w)
		?Just loc = case parseURI loc of
			?None = (Error $ "Redirect URI couldn't be parsed", w)
			?Just uri = doHTTPRequestFollowRedirects {req
				& server_name = maybe loc id uri.uriRegName
				, server_port = maybe 80 id uri.uriPort
				, req_path = uri.uriPath
				, req_query = maybe "" ((+++) "?") uri.uriQuery
				} timeout (maxRedirects-1) w
= (er, w)
